:- module(resource_bounds,
	  [ resource_bounded_call/4	% resource_bounded_call(:Goal, +MaxCPU,
					%			-Status, +Options)
	  ]).
:- use_module(library(option)).
:- use_module(library(debug)).

:- meta_predicate
	resource_bounded_call(0, +, -, +).

/** <module> Resource bounded execution of a goal

This library allows running a goal, providing  limits for the wall time,
CPU time and stack usage.

@see	call_with_time_limit/2
@see	call_with_depth_limit/3
*/

%%	resource_bounded_call(:Goal, +MaxCPU, -Status, +Options) is det.
%
%	Run Goal in a guarded environment,  limiting its resources. Note
%	that resource_bounded_call/4 *|does not guard against mallicious
%	behaviour of Goal!|*
%
%	If all solutions of Goal are  needed,   Goal  must be wrapped in
%	findall/3.
%
%	Below are some examples
%
%	  ==
%	  ?- resource_bounded_call(true, 1, Status, []).
%	  Status = true.
%	  ?- resource_bounded_call(fail, 1, Status, []).
%	  Status = false.
%	  ?- resource_bounded_call((repeat, fail), 0.001, Status, []).
%	  Status = timeout(cpu).
%	  ?- resource_bounded_call(sleep(20), 0.001, Status, [wall_time(1)]).
%	  Status = timeout(wall).
%	  ?- resource_bounded_call(numlist(1, 1000000, L), 1, Status, [global(1000)]).
%	  Status = stack_overflow(global).
%	  ==
%
%	@param	Goal is the goal to execute
%	@param	MaxCPU is the maximum CPU time to use
%	@param	Status is unified with the status.  It is either one of
%		the values returned by thread_join/2 or one of
%
%		  * timeout(cpu)
%		  Goal was aborted due to exceeded CPU Time
%		  * timeout(wall)
%		  Goal was aborted due to exceeded Wall Time
%		  * stack_overflow(global)
%		  Goal terminated due to a global stack overflow
%		  * stack_overflow(local)
%		  Goal terminated due to a local stack overflow
%	@param	Options processed
%
%		  * local(+KCells)
%		  Max local stack usage in cells. Note that the
%		  actual stack limits are rounded up to the nearest
%		  power of two.
%		  * global(+KCells)
%		  Max global stack usage in cells.  The value is
%		  rounded as local(KCells).
%		  * wall_time(Time)
%		  In limit the wall time in addition to the CPU time.

resource_bounded_call(Goal, MaxTime, Status, Options) :-
	queue_options(Options, QueueOpts),
	thread_options(Options, ThreadOpts),
	thread_self(Self),
	term_variables(Goal, Vars),
	thread_create(run(Goal, Vars, Self, Id), Id, ThreadOpts),
	monitor_thread(Goal, Vars, Id, MaxTime, 0, QueueOpts, StatusVar),
	Status = StatusVar.

run(Goal, Vars, Main, Me) :-
	call_cleanup(Goal, thread_send_message(Main, done(Me, Vars))).

queue_options(Options, QueueOpts) :-
	option(wall_time(Time), Options), !,
	get_time(Now),
	Deadline is Time+Now,
	QueueOpts = [deadline(Deadline)].
queue_options(_, []).

thread_options([], []).
thread_options([H0|T0], [H|T]) :-
	thread_option(H0, H), !,
	thread_options(T0, T).
thread_options([_|T0], T) :-
	thread_options(T0, T).

thread_option(local(Cells), local(KB)) :-
	cells_kb(Cells, KB).
thread_option(global(Cells), global(KB)) :-
	cells_kb(Cells, KB).

cells_kb(KCells, KB) :-
	must_be(nonneg, KCells),
	current_prolog_flag(address_bits, Bits),
	KB is KCells * Bits//8.

monitor_thread(Goal, Vars, Id, MaxTime, UsedTime, QueueOpts, Status) :-
	thread_self(Self),
	Left is MaxTime-UsedTime,
	(   thread_get_message(Self, done(Id, Vars),
			       [ timeout(Left)
			       | QueueOpts
			       ])
	->  debug(monitor, 'Done ~w', [Id]),
	    thread_join(Id, Status0),
	    map_status(Status0, Status)
	;   thread_statistics(Id, cputime, CPU),
	    (	(   CPU > MaxTime
		->  Status = timeout(cpu)
		;   QueueOpts = [deadline(Deadline)],
		    get_time(Now),
		    Now > Deadline
		->  Status = timeout(wall)
		)
	    ->  debug(monitor, 'Signalling ~w with abort', [Id]),
	        thread_signal(Id, abort),
		debug(monitor, 'Waiting', []),
		thread_join(Id, _),
		thread_get_message(done(Id, _)),
		debug(monitor, 'Joined', [])
	    ;   monitor_thread(Goal, Vars, Id, MaxTime, CPU, QueueOpts, Status)
	    )
	).

map_status(exception(error(resource_error(stack), Stack)),
	   stack_overflow(Stack)) :- !.
map_status(Status, Status).
